#!/usr/bin/env python3

# Open SSH connections through AWS Session Manager
#
# See https://aws.nz/aws-utils/ec2-ssh for more info.
#
# Author: Michael Ludvig (https://aws.nz)

# The script can list available instances, resolve instance names,
# and host names, etc. In the end it executes 'ssh' with the correct
# parameters to actually start the SSH session.

import os
import sys
import time
import logging
import argparse

from typing import Tuple, List

import botocore.exceptions

from .common import add_general_parameters, show_version, configure_logging, verify_plugin_version
from .resolver import InstanceResolver
from .ec2_instance_connect import EC2InstanceConnectHelper

logger = logging.getLogger("ssm-tools.ec2-ssh")


def parse_args(argv: list) -> Tuple[argparse.Namespace, List[str]]:
    """
    Parse command line arguments.
    """

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter, add_help=False)

    add_general_parameters(parser, long_only=True)

    group_instance = parser.add_argument_group("Instance Selection")
    group_instance.add_argument("--list", dest="list", action="store_true", help="List instances available for SSM Session")

    group_ec2ic = parser.add_argument_group("EC2 Instance Connect")
    group_ec2ic.add_argument("--send-key", dest="send_key", action="store_true", default=True, help="Send the SSH key to instance metadata using EC2 Instance Connect (default and deprecated - use --no-send-key instead)")
    group_ec2ic.add_argument("--no-send-key", dest="send_key", action="store_false", help="Send the SSH key to instance metadata using EC2 Instance Connect")

    parser.description = "Open SSH connection through Session Manager"
    parser.epilog = f"""
IMPORTANT: instances must be registered in AWS Systems Manager (SSM)
before you can start a shell session! Instances not registered in SSM
will not be recognised by {parser.prog} nor show up in --list output.

Visit https://aws.nz/aws-utils/ec2-ssh for more info and usage examples.

Author: Michael Ludvig
"""

    # Parse supplied arguments
    args, extra_args = parser.parse_known_args(argv)

    # If --version do it now and exit
    if args.show_version:
        show_version(args)

    # Require exactly one of INSTANCE or --list
    if bool(extra_args) + bool(args.list) != 1:
        parser.error("Specify either --list or SSH Options including instance name")

    return args, extra_args


def start_ssh_session(ssh_args: list, profile: str = None, region: str = None) -> None:
    aws_args = ""
    if profile:
        aws_args += f"--profile {profile} "
    if region:
        aws_args += f"--region {region} "
    proxy_option = ["-o", f"ProxyCommand=aws {aws_args} ssm start-session --target %h --document-name AWS-StartSSHSession --parameters portNumber=%p"]
    command = ["ssh"] + proxy_option + ssh_args
    logger.debug("Running: %s", command)
    os.execvp(command[0], command)


def main() -> int:
    ## Deprecate old script name
    if sys.argv[0].endswith("/ssm-ssh"):
        print('\033[31;1mWARNING:\033[33;1m "ssm-session" has been renamed to "ec2-session" - please update your scripts.\033[0m', file=sys.stderr)
        time.sleep(3)
        print(file=sys.stderr)

    ## Split command line to main args and optional command to run
    args, extra_args = parse_args(sys.argv[1:])

    if args.log_level == logging.DEBUG:
        extra_args.append("-v")

    configure_logging(args.log_level)

    if not verify_plugin_version("1.1.23", logger):
        sys.exit(1)

    try:
        instance_resolver = InstanceResolver(args)

        if args.list:
            instance_resolver.print_list()
            sys.exit(0)

        # Loop through all SSH args to find:
        # - instance name
        # - user name (for use with --send-key)
        # - key name (for use with --send-key)
        ssh_args = []
        instance_id = ""
        login_name = ""
        key_file_name = ""

        extra_args_iter = iter(extra_args)
        for arg in extra_args_iter:
            # User name argument
            if arg.startswith("-l"):
                ssh_args.append(arg)
                if len(arg) > 2:
                    login_name = arg[2:]
                else:
                    login_name = next(extra_args_iter)
                    ssh_args.append(login_name)
                continue

            # SSH key argument
            if arg.startswith("-i"):
                ssh_args.append(arg)
                if len(arg) > 2:
                    key_file_name = arg[2:]
                else:
                    key_file_name = next(extra_args_iter)
                    ssh_args.append(key_file_name)
                continue

            # If we already have instance id just copy the args
            if instance_id:
                ssh_args.append(arg)
                continue

            # Some args that can't be an instance name
            if arg.startswith("-") or arg.find(":") > -1 or arg.find(os.path.sep) > -1:
                ssh_args.append(arg)
                continue

            # This may be an instance name - try to resolve it
            maybe_login_name = None
            if arg.find("@") > -1:  # username@hostname format
                maybe_login_name, instance = arg.split("@", 1)
            else:
                instance = arg

            instance_id, _ = instance_resolver.resolve_instance(instance)
            if not instance_id:
                # Not resolved as an instance name - put back to args
                ssh_args.append(arg)
                maybe_login_name = None
                continue

            # We got a login name from 'login_name@instance'
            if maybe_login_name:
                login_name = maybe_login_name

            # Woohoo we've got an instance id!
            logger.info("Resolved instance name '%s' to '%s'", instance, instance_id)
            ssh_args.append(instance_id)

            if login_name:
                ssh_args.extend(["-l", login_name])

        if not instance_id:
            logger.warning("Could not resolve Instance ID for '%s'", instance)
            logger.warning("Perhaps the '%s' is not registered in SSM?", instance)
            sys.exit(1)

        if args.send_key:
            EC2InstanceConnectHelper(args).send_ssh_key(instance_id, login_name, key_file_name)

        start_ssh_session(ssh_args=ssh_args, profile=args.profile, region=args.region)

    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        logger.error(e)
        sys.exit(1)

    return 0


if __name__ == "__main__":
    main()



import sys
import logging
import pathlib
import subprocess

# Needed for type hints
import argparse
from typing import List, Tuple, Optional

from .common import AWSSessionBase

logger = logging.getLogger("ssm-tools.ec2-instance-connect")


class EC2InstanceConnectHelper(AWSSessionBase):
    SSH_AGENT_LABEL = "{SSH-AGENT}"

    def __init__(self, args: argparse.Namespace) -> None:
        super().__init__(args)

        # Create boto3 client from session
        self.ec2ic_client = self.session.client("ec2-instance-connect")

    def obtain_ssh_key(self, key_file_name: str) -> Tuple[str, Optional[str]]:
        def _read_ssh_agent_keys() -> List[str]:
            cp = subprocess.run(["ssh-add", "-L"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=False)
            if cp.returncode != 0:
                logger.debug("Failed to run: ssh-add -L: %s", cp.stderr.decode("utf-8").strip().replace("\n", " "))
                return []
            return cp.stdout.decode("utf-8").split("\n")

        def _read_ssh_public_key(key_file_name_pub: str) -> str:
            try:
                key_path = pathlib.Path(key_file_name_pub).expanduser()
                logger.debug("Going to read key from: %s", key_path)
                pub_key_raw = key_path.read_text().split("\n")
                for line in pub_key_raw:
                    if line.startswith("ecdsa-sha2-"):
                        logger.error("ECDSA keys are not yet supported by EC2 Instance Connect: %s", key_file_name_pub)
                        sys.exit(1)
                    if line.startswith("ssh-"):
                        return line
            except (FileNotFoundError, PermissionError) as ex:
                logger.debug("Could not read the public key: %s", ex)
            return ""

        if not key_file_name:
            # No key_file_name was specified, try ...
            # - SSH Agent keys (doesn't matter which one - we'll immediately connect to it)
            ssh_keys = _read_ssh_agent_keys()
            if ssh_keys:
                logger.info("Using SSH key from SSH Agent, should be as good as any.")
                return ssh_keys[0], self.SSH_AGENT_LABEL

            # - ~/.ssh/id_rsa.pub
            ssh_key = _read_ssh_public_key("~/.ssh/id_rsa.pub")
            if ssh_key:
                logger.info("Using SSH key from ~/.ssh/id_rsa.pub - should work in most cases")
                return ssh_key, "~/.ssh/id_rsa"

            # - ~/.ssh/id_dsa.pub
            ssh_key = _read_ssh_public_key("~/.ssh/id_dsa.pub")
            if ssh_key:
                logger.info("Using SSH key from ~/.ssh/id_dsa.pub - should work in most cases")
                return ssh_key, "~/.ssh/id_dsa"

        else:  # i.e. key_file_name is set
            logger.info("Looking for a public key matching: %s", key_file_name)

            # Try reading the public key file (key_file_name + ".pub" suffix)
            key_file_name_pub = key_file_name + ".pub"
            ssh_key = _read_ssh_public_key(key_file_name_pub)
            if ssh_key:
                logger.info("Found a matching SSH Public Key in %s", key_file_name_pub)
                return ssh_key, key_file_name

            # Try reading the public key from SSH Agent
            for line in _read_ssh_agent_keys():
                if line.endswith(key_file_name):
                    logger.info("Found a matching SSH Public Key through SSH Agent")
                    return line, self.SSH_AGENT_LABEL
            logger.debug("Could not find the public key for %s in SSH Agent", key_file_name)

            # Try extracting the public key from the provided private key
            logger.warning("Trying to extract the public key from %s - you may be asked for a passphrase!", key_file_name)
            cp = subprocess.run(["ssh-keygen", "-y", "-f", key_file_name], stdout=subprocess.PIPE, check=False)
            if cp.returncode == 0:
                logger.info("Extracted the public key from: %s", key_file_name)
                return cp.stdout.decode("utf-8").split("\n")[0], key_file_name
            logger.debug("Could not extract the public key from %s", key_file_name)

        logger.warning("Unable to find SSH public key from any available source.")
        logger.warning("Use --debug for more details on what we tried.")
        sys.exit(1)

    def send_ssh_key(self, instance_id: str, login_name: str, key_file_name: str) -> None:
        if not login_name:
            logger.error('Unable to figure out the EC2 login name. Use "-l {user}" or {user}@{instance}.')
            sys.exit(1)

        public_key, private_key_file = self.obtain_ssh_key(key_file_name)
        logger.debug("SSH Key from %s: %s", private_key_file, public_key)

        result = self.ec2ic_client.send_ssh_public_key(
            InstanceId=instance_id,
            InstanceOSUser=login_name,
            SSHPublicKey=public_key,
        )

        if not result["Success"]:
            logger.error("Failed to send SSH Key to %s", instance_id)
            sys.exit(1)



import sys
import pathlib
import logging
import argparse
import subprocess

from typing import List, Tuple

import boto3
import botocore.credentials

import packaging.version
from . import __version__ as ssm_tools_version

__all__ = []

# ---------------------------------------------------------

__all__.append("configure_logging")


def configure_logging(level: int) -> None:
    """
    Configure logging format and level.
    """
    if level == logging.DEBUG:
        logging_format = "[%(name)s] %(levelname)s: %(message)s"
    else:
        logging_format = "%(levelname)s: %(message)s"

    # Default log level is set to WARNING
    logging.basicConfig(level=logging.WARNING, format=logging_format)
    # Except for our modules
    logging.getLogger("ssm-tools").setLevel(level)


# ---------------------------------------------------------

__all__.append("add_general_parameters")


def add_general_parameters(parser: argparse.ArgumentParser, long_only: bool = False) -> argparse._ArgumentGroup:
    """
    Add General Options used by all ssm-* tools.
    """

    # Remove short options if long_only==True
    def _get_opts(opt_long: str, opt_short: str) -> List[str]:
        opts = [opt_long]
        if not long_only:
            opts.append(opt_short)
        return opts

    group_general = parser.add_argument_group("General Options")
    group_general.add_argument(*_get_opts("--profile", "-p"), dest="profile", type=str, help="Configuration profile from ~/.aws/{credentials,config}")
    group_general.add_argument(*_get_opts("--region", "-g"), dest="region", type=str, help="Set / override AWS region.")
    group_general.add_argument(*_get_opts("--verbose", "-v"), action="store_const", dest="log_level", const=logging.INFO, default=logging.INFO, help="Default log level. Show informational messages only.")
    group_general.add_argument(*_get_opts("--debug", "-d"), action="store_const", dest="log_level", const=logging.DEBUG, help="Increase log level.")
    group_general.add_argument(*_get_opts("--quiet", "-q"), action="store_const", dest="log_level", const=logging.WARNING, help="Decrease log level. Only show warnings and errors.")
    group_general.add_argument(*_get_opts("--version", "-V"), action="store_true", dest="show_version", help=f"Show package version and exit. Version is {ssm_tools_version}")
    group_general.add_argument(*_get_opts("--help", "-h"), action="help", help="Print this help and exit")

    return group_general


# ---------------------------------------------------------

__all__.append("show_version")


def show_version(args: argparse.Namespace) -> None:
    """
    Show package version and exit.
    """
    version_string = f"ssm-tools/{ssm_tools_version}"
    if args.log_level <= logging.INFO:
        version_string += f" python/{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
        version_string += f" boto3/{boto3.__version__}"
    print(version_string)
    sys.exit(0)


# ---------------------------------------------------------

__all__.append("bytes_to_human")


def bytes_to_human(size: float) -> Tuple[float, str]:
    """
    Convert Bytes to more readable units
    """
    units = ["B", "kB", "MB", "GB", "TB"]
    unit_idx = 0  # Start with Bytes
    while unit_idx < len(units) - 1:
        if size < 2048:
            break
        size /= 1024.0
        unit_idx += 1
    return size, units[unit_idx]


# ---------------------------------------------------------

__all__.append("seconds_to_human")


def seconds_to_human(seconds: float, decimal: int = 3) -> str:
    """
    Convert seconds to HH:MM:SS[.SSS]

    If decimal==0 only full seconds are used.
    """
    secs = int(seconds)
    fraction = seconds - secs

    mins = int(secs / 60)
    secs = secs % 60

    hours = int(mins / 60)
    mins = mins % 60

    ret = f"{hours:02d}:{mins:02d}:{secs:02d}"
    if decimal:
        fraction = int(fraction * (10**decimal))
        ret += f".{fraction:0{decimal}d}"

    return ret


# ---------------------------------------------------------

__all__.append("verify_plugin_version")


def verify_plugin_version(version_required: str, logger: logging.Logger) -> bool:
    """
    Verify that a session-manager-plugin is installed
    and is of a required version or newer.
    """
    session_manager_plugin = "session-manager-plugin"

    try:
        result = subprocess.run([session_manager_plugin, "--version"], stdout=subprocess.PIPE, check=False)
        plugin_version = result.stdout.decode("ascii").strip()
        logger.debug(f"{session_manager_plugin} version {plugin_version}")

        if packaging.version.parse(plugin_version) >= packaging.version.parse(version_required):
            return True

        logger.error(f"ERROR: session-manager-plugin version {plugin_version} is installed, {version_required} is required")
    except FileNotFoundError:
        logger.error(f"ERROR: {session_manager_plugin} not installed")

    logger.error("ERROR: Check out https://docs.aws.amazon.com/systems-manager/latest/userguide/session-manager-working-with-install-plugin.html for instructions")

    return False


# ---------------------------------------------------------

__all__.append("AWSSessionBase")


class AWSSessionBase:
    def __init__(self, args: argparse.Namespace) -> None:
        # aws-cli compatible MFA cache
        cli_cache = pathlib.Path("~/.aws/cli/cache").expanduser()

        # Construct boto3 session with MFA cache
        self.session = boto3.session.Session(profile_name=args.profile, region_name=args.region)
        self.session._session.get_component("credential_provider").get_provider("assume-role").cache = botocore.credentials.JSONFileCache(cli_cache)



#!/usr/bin/env python3

# Convenience wrapper around 'aws ecs execute-command'
#
# See https://aws.nz/aws-utils/ecs-session for more info.
#
# Author: Michael Ludvig (https://aws.nz)

# The script can list the available containers across all your ECS clusters.
# In the end it executes 'aws ecs execute-command' with the appropriate parameters.
# Supports both EC2 and Fargate ECS tasks.

import os
import sys
import logging
import argparse

from typing import Tuple, List, Dict, Any

import botocore.exceptions

from .common import add_general_parameters, configure_logging, show_version
from .resolver import ContainerResolver

logger = logging.getLogger("ssm-tools.ecs-session")


def parse_args(argv: list) -> Tuple[argparse.Namespace, List[str]]:
    """
    Parse command line arguments.
    """

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter, add_help=False)

    add_general_parameters(parser)

    group_container = parser.add_argument_group("Container Selection")
    group_container.add_argument("CONTAINER", nargs=argparse.ZERO_OR_MORE, help="Task ID, Container Name or IP address. Use multiple keywords (e.g. Task Name and IP) to narrow down ambiguous selections.")
    group_container.add_argument("--list", "-l", dest="list", action="store_true", help="List available containers configured for ECS RunTask")
    group_container.add_argument("--cluster", dest="cluster", metavar="CLUSTER", help="Specify an ECS cluster. (optional)")

    group_session = parser.add_argument_group("Session Parameters")
    group_session.add_argument("--command", dest="command", metavar="COMMAND", default="/bin/sh", help="Command to run inside the container. Default: /bin/sh")

    parser.description = "Execute 'ECS Run Task' in a given container"
    parser.epilog = f"""
IMPORTANT: containers must have "execute-command" setting enabled or they
will not be recognised by {parser.prog} nor show up in --list output.

Visit https://aws.nz/aws-utils/ecs-session for more info and usage examples.

Author: Michael Ludvig
"""

    # Parse supplied arguments
    args, extras = parser.parse_known_args(argv)

    # If --version do it now and exit
    if args.show_version:
        show_version(args)

    # Require exactly one of CONTAINER or --list
    if bool(args.CONTAINER) + bool(args.list) != 1:
        parser.error("Specify either CONTAINER or --list")

    return args, extras


def start_session(container: Dict[str, Any], args: argparse.Namespace, command: str) -> None:
    exec_args = ["aws", "ecs", "execute-command"]
    if args.profile:
        exec_args += ["--profile", args.profile]
    if args.region:
        exec_args += ["--region", args.region]

    # fmt: off
    exec_args += [
        "--cluster", container["cluster_arn"],
        "--task", container["task_arn"],
        "--container", container["container_name"],
        "--command", command,
        "--interactive",
    ]
    # fmt: on

    logger.debug("Running: %s", exec_args)
    os.execvp(exec_args[0], exec_args)


def main() -> int:
    ## Split command line to main args and optional command to run
    args, _ = parse_args(sys.argv[1:])

    configure_logging(args.log_level)

    try:
        if args.list:
            ContainerResolver(args).print_list()
            sys.exit(0)

        container = ContainerResolver(args).resolve_container(keywords=args.CONTAINER)

        if not container:
            logger.warning("Could not find any container matching: %s", " AND ".join(args.CONTAINER))
            sys.exit(1)

        start_session(container, args, args.command)

    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        logger.error(e)
        sys.exit(1)

    return 0


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import sys
import re
import logging
import argparse

from typing import Dict, List, Any, Tuple

import botocore.session

from .common import AWSSessionBase

logger = logging.getLogger("ssm-tools.resolver")


class InstanceResolver(AWSSessionBase):
    def __init__(self, args: argparse.Namespace) -> None:
        super().__init__(args)

        # Create boto3 clients from session
        self.ssm_client = self.session.client("ssm")
        self.ec2_client = self.session.client("ec2")

    def get_list(self) -> Dict[str, Dict[str, Any]]:
        def _try_append(_list: list, _dict: dict, _key: str) -> None:
            if _key in _dict:
                _list.append(_dict[_key])

        items = {}

        # List instances from SSM
        logger.debug("Fetching SSM inventory")
        paginator = self.ssm_client.get_paginator("get_inventory")
        response_iterator = paginator.paginate(
            Filters=[
                {"Key": "AWS:InstanceInformation.ResourceType", "Values": ["EC2Instance", "ManagedInstance"], "Type": "Equal"},
                {"Key": "AWS:InstanceInformation.InstanceStatus", "Values": ["Terminated", "Stopped", "ConnectionLost"], "Type": "NotEqual"},
            ]
        )

        for inventory in response_iterator:
            for entity in inventory["Entities"]:
                logger.debug(entity)
                content = entity["Data"]["AWS:InstanceInformation"]["Content"][0]
                instance_id = content["InstanceId"]
                items[instance_id] = {
                    "InstanceId": instance_id,
                    "InstanceName": "",
                    "HostName": content.get("ComputerName", ""),
                    "Addresses": [content.get("IpAddress")],
                }
                logger.debug("Added instance: %s: %r", instance_id, items[instance_id])

        # Add attributes from EC2
        paginator = self.ec2_client.get_paginator("describe_instances")
        ec2_instance_ids = list(filter(lambda x: x.startswith("i-"), items))

        tries = 5
        while tries:
            # The SSM inventory sometimes returns instances that have been terminated
            # a short while ago which makes the following call fail
            # with InvalidInstanceID.NotFound exception. We'll try and remove the invalid
            # instance ids a {tries} times or until we succeed. If unsuccessful we'll remove
            # the list obtained from SSM without extra details (host name, public IPs, etc).
            # This mostly / only affects accounts with high churn of starting / stopping
            # instances - most users will pass this loop only once.
            try:
                response_iterator = paginator.paginate(InstanceIds=ec2_instance_ids)
                for reservations in response_iterator:
                    for reservation in reservations["Reservations"]:
                        for instance in reservation["Instances"]:
                            instance_id = instance["InstanceId"]
                            if not instance_id in items:
                                continue

                            # Find instance IPs
                            items[instance_id]["Addresses"] = []
                            _try_append(items[instance_id]["Addresses"], instance, "PrivateIpAddress")
                            _try_append(items[instance_id]["Addresses"], instance, "PublicIpAddress")

                            # Store instance AZ - useful for EC2 Instance Connect
                            items[instance_id]["AvailabilityZone"] = instance["Placement"]["AvailabilityZone"]

                            # Find instance name from tag Name
                            for tag in instance.get("Tags", []):
                                if tag["Key"] == "Name" and tag["Value"]:
                                    items[instance_id]["InstanceName"] = tag["Value"]

                            logger.debug("Updated instance: %s: %r", instance_id, items[instance_id])
                    return items

            except botocore.exceptions.ClientError as ex:
                if ex.response.get("Error", {}).get("Code", "") != "InvalidInstanceID.NotFound":
                    raise
                message = ex.response.get("Error", {}).get("Message", "")
                if not message.startswith("The instance ID") or not message.endswith("not exist"):
                    logger.warning("Unexpected InvalidInstanceID.NotFound message: %s", message)
                # Try to extract instace ids ...
                remove_instance_ids = re.findall("i-[0-9a-f]+", message)
                logger.debug("Removing non-existent InstanceIds: %s", remove_instance_ids)
                # Remove the failed ids from the list and try again
                ec2_instance_ids = list(set(ec2_instance_ids) - set(remove_instance_ids))
                tries -= 1

        if not tries:
            logger.warning("Unable to list instance details. Some instance names and IPs may be missing.")

        return items

    def print_list(self) -> None:
        hostname_len = 1  # Minimum of 1 char, otherwise f-string below fails for empty hostnames
        instname_len = 1

        items = self.get_list().values()

        if not items:
            logger.warning("No instances registered in SSM!")
            return

        items_list = list(items)
        del items
        items_list.sort(key=lambda x: x.get("InstanceName") or x.get("HostName"))  # type: ignore

        for item in items_list:
            hostname_len = max(hostname_len, len(item["HostName"]))
            instname_len = max(instname_len, len(item["InstanceName"]))

        for item in items_list:
            print(f"{item['InstanceId']:20}   {item['HostName']:{hostname_len}}   {item['InstanceName']:{instname_len}}   {' '.join(item['Addresses'])}")

    def resolve_instance(self, instance: str) -> Tuple[str, Dict[str, Any]]:
        # Is it a valid Instance ID?
        if re.match("^m?i-[a-f0-9]+$", instance):
            return instance, {}

        # It is not - find it in the list
        instances = []

        items = self.get_list()
        for instance_id in items:
            item = items[instance_id]
            if instance.lower() in [item["HostName"].lower(), item["InstanceName"].lower()] + item["Addresses"]:
                instances.append(instance_id)

        if not instances:
            return "", {}

        if len(instances) > 1:
            logger.warning("Found %d instances for '%s': %s", len(instances), instance, " ".join(instances))
            logger.warning("Use INSTANCE_ID to connect to a specific one")
            sys.exit(1)

        # Found only one instance - return it
        return instances[0], items[instances[0]]


class ContainerResolver(AWSSessionBase):
    def __init__(self, args: argparse.Namespace) -> None:
        super().__init__(args)

        # Create boto3 clients from session
        self.ecs_client = self.session.client("ecs")

        self.args = args
        self.containers: List[Dict[str, Any]] = []
        self._tasks: Dict[str, Any] = {}

    def add_container(self, container: Dict[str, Any]) -> None:
        _task_parsed = container["taskArn"].split(":")[-1].split("/")
        self.containers.append(
            {
                "cluster_name": _task_parsed[1],
                "task_id": _task_parsed[2],
                "cluster_arn": self._tasks[container["taskArn"]]["clusterArn"],
                "task_arn": container["taskArn"],
                "group_name": self._tasks[container["taskArn"]]["group"],
                "container_name": container["name"],
                "container_ip": container["networkInterfaces"][0]["privateIpv4Address"],
            }
        )

    def get_list(self) -> List[Dict[str, Any]]:
        def _try_append(_list: list, _dict: dict, _key: str) -> None:
            if _key in _dict:
                _list.append(_dict[_key])

        # List ECS Clusters
        clusters = []
        logger.debug("Listing ECS Clusters")
        paginator = self.ecs_client.get_paginator("list_clusters")
        for page in paginator.paginate():
            clusters.extend(page["clusterArns"])

        if self.args.cluster:
            filtered_clusters = []
            for cluster in clusters:
                if (self.args.cluster.startswith("arn:") and cluster == self.args.cluster) or cluster.endswith(f"/{self.args.cluster}"):
                    filtered_clusters.append(cluster)
                    break
            clusters = filtered_clusters

        if not clusters:
            logger.warning("No ECS Clusters found.")
            return []

        # List tasks in each cluster
        paginator = self.ecs_client.get_paginator("list_tasks")
        for cluster in clusters:
            logger.debug("Listing tasks in cluster: %s", cluster)

            # maxResults must be <= 100 because describe_tasks() doesn't accept more than that
            for page in paginator.paginate(cluster=cluster, maxResults=100):
                if "taskArns" not in page or not page["taskArns"]:
                    logger.debug(f"No tasks found in cluster {cluster}")
                    break
                response = self.ecs_client.describe_tasks(cluster=cluster, tasks=page["taskArns"])

                # Filter containers that have a running ExecuteCommandAgent
                for task in response["tasks"]:
                    logger.debug(task)
                    self._tasks[task["taskArn"]] = task
                    for container in task["containers"]:
                        if not "managedAgents" in container:
                            continue
                        for agent in container["managedAgents"]:
                            if agent["name"] == "ExecuteCommandAgent" and agent["lastStatus"] == "RUNNING":
                                self.add_container(container)

        return self.containers

    def print_containers(self, containers: List[Dict[str, Any]]) -> None:
        max_len = {}
        for container in containers:
            for key in container.keys():
                if not key in max_len:
                    max_len[key] = len(container[key])
                else:
                    max_len[key] = max(max_len[key], len(container[key]))
        containers.sort(key=lambda x: [x["cluster_name"], x["container_name"]])
        for container in containers:
            print(
                f"{container['cluster_name']:{max_len['cluster_name']}}  {container['group_name']:{max_len['group_name']}}  {container['task_id']:{max_len['task_id']}}  {container['container_name']:{max_len['container_name']}}  {container['container_ip']:{max_len['container_ip']}}"
            )

    def print_list(self) -> None:
        containers = self.get_list()

        if not containers:
            logger.warning("No Execute-Command capable contaianers found!")
            sys.exit(1)

        self.print_containers(containers)

    def resolve_container(self, keywords: List[str]) -> Dict[str, Any]:
        containers = self.get_list()

        if not containers:
            logger.warning("No Execute-Command capable contaianers found!")
            sys.exit(1)

        logger.debug("Searching for containers matching all keywords: %s", " ".join(keywords))

        candidates: List[Dict[str, Any]] = []
        for container in containers:
            for keyword in keywords:
                if keyword not in (container["group_name"], container["task_id"], container["container_name"], container["container_ip"]):
                    logger.debug("IGNORED: Container %s/%s doesn't match keyword: %s", container["task_id"], container["container_name"], keyword)
                    container = {}
                    break
            if container:
                logger.debug("ADDED: Container %s/%s matches all keywords: %s", container["task_id"], container["container_name"], " ".join(keywords))
                candidates.append(container)
        if not candidates:
            logger.warning("No container matches: %s", " AND ".join(keywords))
            sys.exit(1)
        elif len(candidates) == 1:
            self.print_containers(candidates)
            return candidates[0]
        else:
            logger.warning("Found %d instances for: %s", len(candidates), keyword)
            logger.warning("Use Container IP or Task ID to connect to a specific one")
            self.print_containers(candidates)
            sys.exit(1)



import time
import logging
import pexpect

logger = logging.getLogger("ssm-tools.talker")


class SsmTalker:
    def __init__(self, instance_id: str, profile: str, region: str) -> None:
        self._instance_id = instance_id
        self.connect(instance_id, profile, region)

    def connect(self, instance_id: str, profile: str, region: str) -> None:
        extra_args = ""
        if profile:
            extra_args += f"--profile {profile} "
        if region:
            extra_args += f"--region {region} "
        command = f"aws {extra_args} ssm start-session --target {instance_id}"
        logger.debug("Spawning: %s", command)
        self._child = pexpect.spawn(command, echo=False, encoding="utf-8", timeout=10)
        logger.debug("PID: %s", self._child.pid)

        self.wait_for_prompt()
        logger.debug(self._child.before.strip())
        self.shell_prompt = self._child.after

        # Turn off input echo
        self._child.sendline("stty -echo")
        self.wait_for_prompt()

        # Change to home directory (SSM session starts in '/')
        self._child.sendline("cd")
        self.wait_for_prompt()

    def exit(self) -> None:
        logger.debug("Closing session")
        self._child.sendcontrol("c")
        time.sleep(0.5)
        self._child.sendline("exit")
        try:
            self._child.expect(["Exiting session", pexpect.EOF])
        except (OSError, pexpect.exceptions.EOF):
            pass

    def wait_for_prompt(self) -> None:
        """
        As of now a typical SSM prompt is 'sh-4.2$ '
        """
        self._child.expect(".*\$ $")



# aws-ssm-tools packaging

import os
import sys
import pathlib

from setuptools import setup, find_packages
from setuptools.command.install import install

import ssm_tools

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

SCRIPTS = [
    "ec2-session",
    "ecs-session",
    "ec2-ssh",
    "ssm-tunnel",
    # Renamed, deprecated and soon to be removed...
    "ssm-session",
    "ssm-ssh",
]
VERSION = ssm_tools.__version__

requirements = HERE / "requirements.txt"
with requirements.open() as f:
    reqs = [req.strip() for req in f.readlines() if req.strip() and not req.startswith("#")]


class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""

    description = "verify that the git tag matches our version"

    def run(self) -> None:
        tag = os.getenv("CIRCLE_TAG")
        if not tag:
            sys.exit("Env var $CIRCLE_TAG is not defined - are we running a CircleCI build?")

        if tag.startswith("v"):  # If tag is v1.2.3 make it 1.2.3
            tag = tag[1:]

        if tag != VERSION:
            info = f"Git tag: {tag} does not match the version of this app: {VERSION}"
            sys.exit(info)


def console_scripts() -> list:
    scripts = []
    for script in SCRIPTS:
        # All script entries must be in this format:
        # "ec2-session = ssm_tools.ec2_session_cli:main"
        if script in ["ec2-session", "ec2-ssh"]:
            scripts.append(f"{script} = ssm_tools.{script.replace('-','_').replace('ec2','ssm')}_cli:main")
        else:
            scripts.append(f"{script} = ssm_tools.{script.replace('-','_')}_cli:main")
    return scripts


setup(
    name="aws-ssm-tools",
    version=VERSION,
    packages=find_packages(),
    entry_points={
        "console_scripts": console_scripts(),
    },
    python_requires=">=3.6",
    install_requires=reqs,
    package_data={
        "": ["*.txt", "*.md", "ssm-tunnel-updown.dns-example", "LICENSE"],
    },
    author="Michael Ludvig",
    author_email="mludvig@logix.net.nz",
    description="Tools for AWS Systems Manager: " + " ".join(SCRIPTS),
    long_description=README,
    long_description_content_type="text/markdown",
    license="Apache License 2.0",
    keywords="aws ssm " + " ".join(SCRIPTS),
    url="https://github.com/mludvig/aws-ssm-tools",
    project_urls={
        "Bug Tracker": "https://github.com/mludvig/aws-ssm-tools/issues",
        "Documentation": "https://github.com/mludvig/aws-ssm-tools/blob/master/README.md",
        "Source Code": "https://github.com/mludvig/aws-ssm-tools",
    },
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Apache Software License",
        "Development Status :: 5 - Production/Stable",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: System :: Systems Administration",
        "Topic :: System :: Networking",
    ],
    cmdclass={
        "verify": VerifyVersionCommand,
    },
)



#!/usr/bin/env python3

# Convenience wrapper around 'aws ssm start-session'
# can resolve instance id from Name tag, hostname, IP address, etc.
#
# See https://aws.nz/aws-utils/ssm-session for more info.
#
# Author: Michael Ludvig (https://aws.nz)

# The script can list available instances, resolve instance names,
# and host names, etc. In the end it executes 'aws' to actually
# start the session.

import os
import sys
import time
import logging
import signal
import argparse

from typing import Tuple, List

import botocore.exceptions

from .common import add_general_parameters, show_version, configure_logging
from .resolver import InstanceResolver

logger = logging.getLogger("ssm-tools.ec2-session")

# Only on Linux / Unix / Mac because Windows don't interpret Ctrl-Z
if os.name == "posix":
    # Ignore Ctrl-Z - pass it to the shell
    signal.signal(signal.SIGTSTP, signal.SIG_IGN)


def parse_args(argv: list) -> argparse.Namespace:
    """
    Parse command line arguments.
    """

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter, add_help=False)

    add_general_parameters(parser)

    # fmt: off
    group_instance = parser.add_argument_group("Instance Selection")
    group_instance.add_argument("INSTANCE", nargs="?", help="Instance ID, Name, Host name or IP address")
    group_instance.add_argument("--list", "-l", dest="list", action="store_true", help="List instances available for SSM Session")

    group_session = parser.add_argument_group("Session Parameters")
    group_session.add_argument("--user", "-u", "--sudo", dest="user", metavar="USER", help="SUDO to USER after opening the session. Can't be used together with --document-name / --parameters. (optional)")
    group_session.add_argument("--command", "-c", dest="command", metavar="COMMAND", help="Command to run in the SSM Session. Can't be used together with --user. "
                               "If you need to run the COMMAND as a different USER prepend the command with the appropriate 'sudo -u USER ...'. (optional)")
    group_session.add_argument("--document-name", dest="document_name", help="Document to execute, e.g. AWS-StartInteractiveCommand (optional)")
    group_session.add_argument("--parameters", dest="parameters", help="Parameters for the --document-name, e.g. 'command=[\"sudo -i -u ec2-user\"]' (optional)")
    # fmt: on

    parser.description = "Start SSM Shell Session to an EC2 instance"
    parser.epilog = f"""
IMPORTANT: instances must be registered in AWS Systems Manager (SSM)
before you can start a shell session! Instances not registered in SSM
will not be recognised by {parser.prog} nor show up in --list output.

Visit https://aws.nz/aws-utils/ssm-session for more info and usage examples.

Author: Michael Ludvig
"""

    # Parse supplied arguments
    args = parser.parse_args(argv)

    # If --version do it now and exit
    if args.show_version:
        show_version(args)

    # Require exactly one of INSTANCE or --list
    if bool(args.INSTANCE) + bool(args.list) != 1:
        parser.error("Specify either INSTANCE or --list")

    if args.parameters and not args.document_name:
        parser.error("--parameters can only be used together with --document-name")

    if bool(args.user) + bool(args.command) + bool(args.document_name) > 1:
        parser.error(
            """
Use only one of --user / --command / --document-name
If you need to run the COMMAND as a specific USER then prepend
the command with the appropriate: sudo -i -u USER COMMAND
"""
        )

    return args


def start_session(instance_id: str, args: argparse.Namespace) -> None:
    exec_args = ["aws", "ssm", "start-session"]
    if args.profile:
        exec_args += ["--profile", args.profile]
    if args.region:
        exec_args += ["--region", args.region]

    if args.user:
        # Fake --document-name / --parameters for --user
        exec_args += ["--document-name", "AWS-StartInteractiveCommand", "--parameters", f'command=["sudo -i -u {args.user}"]']
    if args.command:
        # Fake --document-name / --parameters for --command
        exec_args += ["--document-name", "AWS-StartInteractiveCommand", "--parameters", f"command={args.command}"]
    else:
        # Or use the provided values
        if args.document_name:
            exec_args += ["--document-name", args.document_name]
        if args.parameters:
            exec_args += ["--parameters", args.parameters]

    exec_args += ["--target", instance_id]
    logger.debug("Running: %s", exec_args)
    os.execvp(exec_args[0], exec_args)


def main() -> int:
    ## Deprecate old script name
    if sys.argv[0].endswith("/ssm-session"):
        print('\033[31;1mWARNING:\033[33;1m "ssm-session" has been renamed to "ec2-session" - please update your scripts.\033[0m', file=sys.stderr)
        time.sleep(3)
        print(file=sys.stderr)

    ## Split command line to main args and optional command to run
    args = parse_args(sys.argv[1:])

    configure_logging(args.log_level)

    try:
        if args.list:
            InstanceResolver(args).print_list()
            sys.exit(0)

        instance_id, _ = InstanceResolver(args).resolve_instance(args.INSTANCE)

        if not instance_id:
            logger.warning("Could not resolve Instance ID for '%s'", args.INSTANCE)
            logger.warning("Perhaps the '%s' is not registered in SSM?", args.INSTANCE)
            sys.exit(1)

        start_session(instance_id, args)

    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        logger.error(e)
        sys.exit(1)

    return 0


if __name__ == "__main__":
    main()
